
PRO freefree_k, kappa, temp, wvl, int, sumt=sumt, dem_int=dem_int, $
              photons=photons, min_abund=min_abund, verbose=verbose, $
              kev=kev, em_int=em_int, sngl_ion=sngl_ion, element=element, $
              abund_file=abund_file, ioneq_file=ioneq_file


;+
; NAME:
;     FREEFREE
;
; PURPOSE:
;     Computes the free-free (bremsstrahlung) continuum spectrum for
;     an ionized plasma.
;
; CATEGORY:
;     CHIANTI; continuum.
;
; CALLING SEQUENCE:
;     FREEFREE, Temp, Wvl, Int
;
; INPUTS:
;     Temp:   Temperature (in K). Can be an array. If the input
;             DEM_INT= is used then the temperatures must be spaced at
;             equal intervals of 
;             log10(T). E.g., log T=6.0, 6.1, 6.2, 6.3, ...
;
;     Wvl:    Wavelengths in angstroms. Can be a scalar or vector. If /keV is
;             set, then WVL is assumed to contain energies in keV units.
;
; OPTIONAL INPUTS:
;     Dem_Int:  This should be the same size as TEMP and contain
;               differential emission measure values. Specifying
;               DEM_INT places a restriction on TEMP (see above). Note
;               that DEM_INT values are multiplied by the factor dT
;               within the routine, whereas EM_INT values are not.
;
;     Em_Int:   This should be the same size as TEMP and contain
;               emission measure values. Note
;               that DEM_INT values are multiplied by the factor dT
;               within the routine, whereas EM_INT values are not. If
;               neither EM_INT or DEM_INT are specified, then a EM_INT
;               vector of same size as TEMP with values of unity is
;               assumed.
;
;     Min_Abund:This keyword allows the specification of a minimum abundance, 
;               such that any elements with an abundance (relative to 
;               hydrogen) less than MIN_ABUND will not be included in the 
;               calculation. E.g., MIN_ABUND=1e-5.
;
;     Abund_File: The name of an element abundance file to use. If not
;                 specified, then a widget will appear allowing you to
;                 choose one of CHIANTI abundance files.
;
;     Ioneq_File: The name of an ionization equilibrium file to use. If
;                 not specified, then the default CHIANTI ionization
;                 file (!ioneq_file) is used.
;
;     Sngl_Ion: A string containing the name of an ion in CHIANTI
;               format (e.g., 'fe_21' for Fe XXI). Can be an array. The
;               continuum will only be calculated for the specified
;               ions. 
;
;     Element:  Either the atomic number of an element, or the
;               symbol name (e.g., 'Fe' for iron). If set, then the
;               continuum will only be computed for the specified
;               element. This can not be an array.
;	
; KEYWORD PARAMETERS:
;     SUMT:    The default is to output the intensity array as an array 
;              of size (nwvl x nT). Setting this keyword performs a sum 
;              over the temperatures to yield a vector of same size as 
;              the input wavelengths, thus producing the complete 
;              free-free spectrum.
;
;     PHOTONS: Gives output emissivity in photon units rather than ergs.
;
;     KEV:     If set, then WVL is assumed to contain energies in keV units
;              rather than wavelengths in Angstroms.
;
;     VERBOSE: If set, then information will be printed to the screen.
;
; OUTPUTS:
;     Int:    Free-free continuum intensity in units of 
;             10^-40 erg cm^3/s/sr/Angstrom  
;             [ integral(N_H N_e dh) in cm^-5 ] if a DEM is not defined. 
;
;             If DEM values are defined, it is assumed that they are given
;             as N_H N_e dh/dT.  The units are 10^-40 erg/cm^2/s/sr/Angstrom. 
;
;             If T is given as a 1-D array, then the output will be a 2-D array,
;             with one element for each temperature and wavelength 
;             (but also see SUMT).
;
;             If the keyword /keV is set, then the units of INT will be 
;             10^-40 erg cm^3/s/sr/keV
;
; CALLS:
;     SUTHERLAND, ITOH
;
; PROGRAMMING NOTES:
;     This routine calls out to sutherland.pro and itoh.pro to
;     actually calculate the continuum. 
;
;     The Itoh fitting formula is only valid for (6.0 LE logT LE 8.5). 
;     For temperatures below this, we thus switch to the Sutherland 
;     fitting formula. There is very little (<1%) between the two at 
;     logT=6.
;
;     Itoh also has a constraint on the quantity u=hc/kTl (l=wavelength), 
;     such that (-4 LE log u LE 1.0). The upper limit corresponds to the 
;     continuum being cut-off prematurely at low wavelengths. E.g., for 
;     T=10^6 the cutoff is at 14.39 angstroms. For these low wavelengths 
;     we also use the Sutherland data to complete the continuum. Note that 
;     the continuum at these wavelengths is very weak
;
;     To calculate the continuum spectrum the user should specify
;     whether the plasma structure is specified through a differential
;     emission measure (DEM) curve or through emission measure (EM)
;     values. For the former, the DEM values are specified through the
;     DEM_INT keyword and for the latter the EM values are specified
;     through the EM_INT keyword. If neither are specified EM_INT
;     values of unity are assumed.
;
;     If DEM_INT values are specifed they must be specified on a
;     temperature scale that has constant intervals in log10(T). EM_INT
;     values do not have this restriction.
;
;     Note that EM_INT must be used for isothermal plasmas.
;
; EXAMPLE:
;     IDL> abfile=ch_choose_abund()
;     IDL> wvl=findgen(100)+1.
;     IDL> freefree,1e7,wvl,int,abund_file=abfile
;     IDL> freefree,1e7,wvl,int,abund_file=abfile,element='fe'
;
; MODIFICATION HISTORY:
;
;    Ver.1, 11-Jul-2014, Jaroslav Dudik
;         written
;	    14-Oct-2014, Jaroslav Dudik
;	  added no_setup keyword
;         disabled checks for keyword_set(abund) and keyword_set(abund_name)
;	    18-Oct-2014, Jaroslav Dudik
;         added /keV, according to freebound_k.pro (should be fine)
;	  added /dem_int
; 	  added /sumt
;	
;            05-Oct-2015, Jaroslav Dudik
;	  added em_int=em_int
; 
;    Ver.2, 2-Mar-2021,E. Dz.
;         for KAPPA based on CH10
;
;	2023-11-13	JD	commented out no_setup; should not be needed
;	2023-11-15	JD	changed to interpolation in alog10(int_precalc)
;


IF n_params() lt 4 then begin
  print,'Use: IDL> freefree_k, kappa, temp, wvl, int [, /sumt, dem_int=, em_int= '
   print,'                    /photons, min_abund=, /verbose, /kev, '
  print,'                    sngl_ion=, element=, ioneq_file='
  print,'                    abund_file= ] '
  print,'  '
  return
ENDIF 


t1=systime(1)

temp=double(temp)
wvl=double(wvl)
;kp		= ['1.55','1.6','1.7','1.8','02', '03', '04', '05', '07', '10', '15', '25', '33']
kp		= ['02', '03', '04', '05', '07', '10', '15', '25', '33']
ik		= where(kappa EQ double(kp))
;s_k             = ['01p55','01p60','01p70','01p80','02', '03', '04', '05', '07', '10', '15', '25', '33']
s_k             = ['02', '03', '04', '05', '07', '10', '15', '25', '33']

if (ik EQ -1) then begin
  print, '%FREEFREE_K: Warning:  Incorrect kappa value!'
  stop
endif else distr = s_k[ik]
;
; * keV units *
; For the calculation of the continuum, it's simply necessary to pass the
; /kev keyword on to itoh.pro and sutherland.pro. However, there's also a
; check in freefree.pro based on wavelength that needs to be corrected. I
; thus define 'chckwvl' for this, with a different definition in the case
; of /kev.
; 
IF keyword_set(kev) THEN BEGIN
  wvl_save=wvl
  efact=12.398/wvl^2   ; hc/E^2 factor to multiply continuum by
  efact=reverse(efact)
  wvl=12.398/wvl   ; convert energies to angstroms
  wvl=reverse(wvl)
ENDIF ELSE BEGIN
  efact=1d0
ENDELSE


IF n_elements(dem_int) NE 0 THEN BEGIN
   IF n_elements(dem_int) NE n_elements(temp) THEN BEGIN
      print,'% FREEFREE_K: Warning, number of elements of DEM_INT must match'
      print,'  the number of temperatures. Freefree continuum not calculated.'
      return
   ENDIF
ENDIF

IF n_elements(abund_file) NE 0 THEN BEGIN
  chck=file_search(abund_file,count=count)
  IF count EQ 0 THEN BEGIN
    print,'% FREEFREE_K: the specified abund_file was not found. Returning...'
    return
  ENDIF 
ENDIF ELSE BEGIN
  abund_file=''
ENDELSE 
read_abund,abund_file,abund,ab_ref, element=element

if NOT(keyword_set(min_abund))	 then min_abund   = 0d
  NZ=n_elements(abund) 
  iabund	= where(abund[0:NZ-1] Gt min_abund)


IF n_elements(ioneq_file) EQ 0 THEN BEGIN
  ioneq_file=!data_pth+'ioneq/kappa_'+kp(ik)+'.ioneq'
ENDIF ELSE BEGIN
  chck=file_search(ioneq_file,count=count)
  IF count EQ 0 THEN BEGIN
    print,'% FREEFREE_K: the specified ioneq_file was not found. Returning...'
    return
  ENDIF 
ENDELSE

; This converts the input SNGL_ION (specifying the recombined ions) to
; the recombining ions.
;
IF n_elements(sngl_ion) NE 0 THEN BEGIN
  nion=n_elements(sngl_ion)
  recomb_ion=strarr(nion)
  FOR i=0,nion-1 DO BEGIN
    convertname,sngl_ion[i],a,b
    zion2name,a,b+1,rec_ion
    recomb_ion[i]=trim(rec_ion)
  ENDFOR 
ENDIF


;read the file
read_ff_k, !data_pth +'/continuum/ff_kappa_'+kp[ik]+'.dat', NZ, wvl_precalc, temp_precalc, ff;, ref
NWVL_precalc	= n_elements(wvl_precalc)
NT_precalc	= n_elements(temp_precalc)
int_precalc	= dblarr(NWVL_precalc,NT_precalc)
NT		= n_elements(temp)
NWVL		= n_elements(wvl)
int		= dblarr(NWVL,NT)
int1		= dblarr(NWVL_precalc,NT)
ident_wvl	= make_array(nwvl,val=1d)
ident_t		= make_array(nt,val=1d)



;constants [cgs]
elcharge	= 1.6021764D-19 *2997924580D					; [statcoulomb]
h		= 6.626069D-27							; [erg.s]
me		= 9.109382D-28							; [g]
c		= 2.99793D+10							; [cm/s]
k_B		= 1.38065D-16							; [erg/K]
C_factor	= 32D*!dpi/3D *sqrt(2D*!dpi/3D) *elcharge^6D /me^(3/2D) /c^3 /k_B^(1/2D)
Ry		= 2D*!dpi^2D *elcharge^4D *me / h^2D				; Rydberg energy [erg]

; IF NOT keyword_set(no_setup) THEN BEGIN
;    setup_elements
;    no_setup=1
; ENDIF

;assemble the original intensity using the provided abundances
for iw=0, NWVL_precalc-1, 1 do begin
 for it=0, NT_precalc-1, 1 do begin
  for iz=0, n_elements(iabund)-1, 1 do begin

	int_precalc[iw,it] = int_precalc[iw,it] +abund[iabund[iz]] *ff[iabund[iz],iw,it]
  endfor
 endfor
endfor

;2. calculation of the output - int[iwvl,it]
;-------------------------------------------------
;interpolation in log(T), original wavelengths
for il=0, NWVL_precalc-1, 1 do begin
  ; int1[il,*] = interpol(reform(int_precalc[il,*]), $
		; 	alog10(temp_precalc[*]), alog10(temp[*]))
  int1[il,*] = 10.^interpol(reform(alog10(int_precalc[il,*])), $
			alog10(temp_precalc[*]), alog10(temp[*]))
endfor

;spline-interpolation in wavelength
for it=0, NT-1, 1 do begin
  ; int[*,it] = spl_interp(alog10(wvl_precalc[*]), int1[*,it], $
		; 	 spl_init(alog10(wvl_precalc[*]), int1[*,it]), alog10(wvl[*]))
  int[*,it] = 10.^spl_interp(alog10(wvl_precalc[*]), alog10(int1[*,it]), $
			 spl_init(alog10(wvl_precalc[*]), alog10(int1[*,it])), alog10(wvl[*]))
endfor

;multiply by 1d40
int	= int *1d40


; ;DEM_INT
; ; IF n_elements(dem_int) NE 0 THEN BEGIN
; ;   demfactor=temp*0.1d/alog10(exp(1d))
; ;   int=int* (ident_wvl # (dem_int*demfactor))
; ; ENDIF

;
; DEM_INT and EM_INT implementation
; ---------------------------------
;  If DEM_INT is specified then it's necessary to multiply by the
;  DEM values by the factor dT (=ln(10)*T*d(log10(T)).
;
;  I implement this by introducing a keyword dT. If dT=1 then a DEM is
;  assumed, while dt=0 implies EM values. In the latter case DEM_INT
;  is set to be EM_INT.
;
;  EM_VALS is set to be DEM_INT or EM_INT and is used in the rest of
;  the code.
;
CASE 1 OF 
  n_elements(dem_int) NE 0 AND n_elements(em_int) NE 0: BEGIN
    print,'%FREEFREE_K: Please specify only one of DEM_INT and EM_INT. Returning...'
    int=0.
    return
  END
 ;
  n_elements(dem_int) EQ 0 AND n_elements(em_int) NE 0: BEGIN
    em_vals=em_int
    dt=0
  END 
 ;
  n_elements(dem_int) NE 0 AND n_elements(em_int) EQ 0: BEGIN
    em_vals=dem_int
    dt=1
  END 
 ;
  n_elements(dem_int) EQ 0 AND n_elements(em_int) EQ 0: BEGIN
    em_vals=dblarr(nt)+1.
    dt=0
  END 
ENDCASE

;
; Make sure EM_VALS has same size as TEMP
;
em_vals=double(em_vals)
IF n_elements(em_vals) NE nt THEN BEGIN
  print,'%FREEFREE_K: Warning, number of elements of DEM_INT (EM_INT) must match the'
  print,'  number of temperatures. Returning...'
  int=0.
  return
ENDIF 


;
; If /DT set then need to multiply by dT for each element of EM_VALS.
;
IF keyword_set(dt) THEN BEGIN
  IF nt LT 2 THEN BEGIN
    print,'%FREEFREE_K: Two or more temperatures must be specified if DEM_INT is specified.'
    print,'              Returning...'
    int=0.
    return
  ENDIF 
  ;
  ;
  lt1=alog10(temp[0:nt-2])
  lt2=alog10(temp[1:*])
  mean_lt=mean(lt2-lt1)
  max_LT=max(lt2-lt1)
  min_LT=min(lt2-lt1)
  IF max_LT-mean_LT GE 0.05*mean_LT OR mean_LT-min_LT GE 0.05*mean_LT THEN BEGIN
    print,'%FREEFREE_K: Error, LOG10 TEMP needs to be evenly spaced if DEM_INT is specified.'
    print,'              E.g., LOG10 TEMP should be spaced at 0.1 dex intervals.'
    int=0.
    return
  ENDIF 
  demfactor=temp*mean_lt/alog10(exp(1))
ENDIF ELSE BEGIN
  demfactor=dblarr(nt)+1.
ENDELSE 
;
int=int* (ident_wvl # (em_vals*demfactor))

;ergs
if NOT(keyword_set(photons)) then begin
  int	= int *((h*c/wvl[*]/1d-8)#ident_t)
endif

;multiply int by the hc/E^2 factor if /kev set
if keyword_set(kev) then begin
  e_arr=efact#ident_t
  int=int*e_arr
 ;
 ; reverse the wavelength dimension to match energy ordering rather than
 ; wavelength ordering
  int=int[reverse(indgen(nwvl)),*]
 ;
 ; and set wavelengths back to energy units
  wvl=wvl_save 

endif

siz=size(int)

IF keyword_set(sumt) AND (siz[0] GT 1) THEN int=total(int,2)

IF keyword_set(verbose) THEN BEGIN 
  t2=systime(1)
  print,format='("% FREEFREE_K: continuum calculated in ",f8.1," seconds")',t2-t1
ENDIF  

END
