pro FREEFREE_k_INTEGRAL, kappa, temp, wvl, int,  	min_ionfrac=min_ionfrac, min_abund=min_abund, $
							abund_name=abund_name, iabund=iabund, photons=photons, $
							ff=ff


; NAME
;
;     FREEFREE_K_INTEGRAL
;
;  PROJECT
;
;     Modified CHIANTI 7.1 for kappa-distributions
;
;  PURPOSE:
;
;     This routine computes the free-free continuum (bremsstrahlung) 
;     using the Eq. 4 of Dudik et al. (A&A 539, A107)
;
;     The integration is carried out directly using the gaunt factors of
;     Sutherland (MNRAS 300, 321, 1998)
;
;   NOTES:
;
;      The integration takes a lot of time if a range of temperatures
;      and/or wavelengths and/or ions is specified.
;      Since the Sutherland GFF_EW values are scaled, the integration
;      has to be performed over each ion separately and then summed
;      to produce the spectrum.
;
;      To reduce the integration time by dropping insignificant ions,
;      following restrictions can be imposed (see OPTIONAL INPUTS)
;      abundances              > min_abund 
;      ion(ionization energy)  > 0
;      ion fraction (Temp)     > min_ionfrac
;
;      The integration is carried in CGS units. All variables and constants
;      in this program are in CGS units.
;
;
; INPUTS
;
;    KAPPA   Value of kappa. Single scalar equal to 2, 3, 4, 5, 6, 10, 
;                            15, 25, or 33. Other values do not work.
;                                           (no ioneq files)
;
;    TEMP    Temperature (in K). Can be an array. If the input
;            DEM_INT= is used then the temperatures must be spaced at
;            equal intervals of 
;            log10(T). E.g., log T=6.0, 6.1, 6.2, 6.3, ...
;
;    WVL     Wavelengths in angstroms. Can be a scalar or vector.
;            Wavelengths should be between 0.1A and 1m.
;
;
; OUTPUTS
;
;    INT     Free-free continuum intensity in units of 
;            erg cm^3/s/sr/Angstrom  
;
; OPTIONAL INPUTS
;
;    MIN_ABUND This keyword allows the specification of a minimum abundance, 
;              such that any elements with an abundance (relative to 
;              hydrogen) less than MIN_ABUND will not be included in the 
;              calculation. E.g., MIN_ABUND=1e-5.
;              Default value is 1d-6, which saves about 1/2 of the calc time.
;
;    MIN_IONFRAC  minimum ionization fraction to take into account.
;                 Default is 1d-4
;
; OPTIONAL OUTPUTS
;
;    FF        (keyword) free-free continuum as a function of Z
;
;
; KEYWORDS
;
;    ABUND_NAME Name of the abundance file
;
;    IABUND   Vector containing the Z-1 of elements to be included
;             in the calculation
;
;    PHOTONS  Gives output emissivity in photon units rather than ergs.
;
;    FF       (named variable) contributions to free-free continuum 
;             from individual elements. Has to be folded over abundances
;             to obtain the final free-free spectrum.
;
; CALLS
;
;    READ_GFFEW (appended)
;
; COMMON BLOCKS
;
;    NONE (we try to avoid these)
;
;
; EXAMPLES
;
;       IDL> freefree_k_integral, 3, 1d7, wvl, int
;       IDL> freefree_k_integral, 3, 1d7, wvl, int, min_abund=1d-6
;       IDL> freefree_k_integral, 3, 1d7, wvl, int, min_abund=1d-6, min_ionfrac=1d-3
;
;       IDL> wvl=findgen(5001)/10. + 50.
;       IDL> temp=10.^(findgen(41)/10. +4.)
;       IDL> freefree_k_integral, 3, temp, wvl, int
;
;
; MODIFICATION HISTORY
;
;    Ver.1, 10-Jul-2014, Jaroslav Dudik
;         Adapted from ff_spectrum_k_Z.pro
;
;
; 2023-04-24	JD		Changed for new density-dependent files
;				(this concerns mostly the 'kp' variable input)
; 2023-11-18	JD		moved to chianti10.2_kappa/


t0 = systime(1)

;1. Setup
;--------
kp		= ['1.7', '1.8', '1.9', '2', '2.5', '3', '4', '5', '7', '10', '15', '25', '33']
kp_filename	= ['1p7', '1p8', '1p9', '2', '2p5', '3', '4', '5', '7', '10', '15', '25', '33']
ik		= where(kappa EQ double(kp))
Akappa		= Gamma(double(kappa)+1D)/Gamma(double(kappa) -1/2D) $
		  /(double(kappa)-3/2D)^1.5D					; normalisation constant

if (ik EQ -1) then begin
  message, 'Incorrect kappa value!'
  stop
endif
distr		= kp[ik]
;read_ioneq, !data_pth +'/ioneq/kappa_'+kp[ik]+'.ioneq', t_ioneq, ioneq, ioneq_ref
; read_ioneq, !data_pth +'ioneq_dens/Dz23_logn0_kappa_'+kp_filename[ik]+'.ioneq', t_ioneq, ioneq, ioneq_ref	; 2023-04-24
read_ioneq, !data_pth +'/ioneq/Dz23_logn0_kappa_'+kp_filename[ik]+'.ioneq', t_ioneq, ioneq, ioneq_ref		; 2023-11-18

wvl		= wvl *1D-8							; Angstrom --> cm
NWVL		= n_elements(wvl)
NT		= n_elements(temp)

if (min(wvl) LE 0.99d-9) or (max(wvl) GT 1.01d2) then begin
  message, 'Wavelength outside the 0.1A - 1m range'
  stop
endif


;constants [cgs]
elcharge	= 1.6021764D-19 *2997924580D					; [statcoulomb]
h		= 6.626069D-27							; [erg.s]
me		= 9.109382D-28							; [g]
c		= 2.99793D+10							; [cm/s]
k_B		= 1.38065D-16							; [erg/K]

C_factor	= 32D*!dpi/3D *sqrt(2D*!dpi/3D) *elcharge^6D /me^(3/2D) /c^3 /k_B^(1/2D)
Ry		= 2D*!dpi^2D *elcharge^4D *me / h^2D				; Rydberg energy [erg]


;abundances
if NOT(keyword_set(min_ionfrac)) then min_ionfrac = 1d-4
if NOT(keyword_set(min_abund))	 then min_abund   = 1d-6
if NOT(keyword_set(abund_name))	 then read_abund, '', abund, ref_abund else read_abund, !xuvtop+ '/abundance/' +abund_name, abund, ref_abund
if NOT(keyword_set(iabund)) then begin
  NZ		= 30
; -!- There are 50 elements in abundance files, but only 30 ionization potentials in the .ionpot file.
  iabund	= where(abund[0:NZ-1] GE min_abund)
endif

int		= dblarr(NWVL,NT)
ff	 	= dblarr(NZ,NWVL,NT)



;2. free-free continuum computation
;----------------------------------
read_gffew, 	gffew, ei, wi							; read the free-free gaunt factor
read_ip, 	!xuvtop+'/ip/chianti.ip', ionpot, ipref				; ionization potentials [cm^-1]

NY		= 2001
y		= dindgen(NY)/(NY-1) *40D *10D
z		= dblarr(NWVL,NT)
integrand	= dblarr(NY)



;temperature FOR
for it = 0, NT-1, 1 do begin
   
  ;abundance FOR
  for iz=0, n_elements(iabund)-1, 1 do begin

    iions	= where(ionpot(iabund[iz],*) GT 0D)

    ;ionization stages FOR
    for ii=0, n_elements(iions)-1, 1 do begin

	ion_fraction	= interpol(ioneq[*,iabund[iz],iions[ii]+1], 10^t_ioneq, temp[it])

	if (ion_fraction GT min_ionfrac) then begin

	  for iw = 0, NWVL-1, 1 do begin

		;scaling factor
		scf			= ionpot(iabund[iz],iions[ii])*Ry/( 109737.3*k_B*temp[it] )
		z[iw,it]		= h*c/(wvl[iw] *k_B*temp[it])
		for q = 0, NY-1, 1 do begin
		  integrand[q]		= interpolate(gffew, 3D*(alog10(y[q]/scf) +9D), 3D*(alog10(z[iw,it]/scf) +8D)) $
					  / ( 1D +(y[q]+z[iw,it])/(double(kappa)-3/2D) )^(double(kappa)+1D)
		endfor
	
		;in [phot.cm^3/s/sr/A]
		ff[iabund[iz],iw,it]	= ion_fraction *(iions[ii]+1)^2D *C_factor *Akappa /sqrt(temp[it]) 	$
								*int_tabulated(y, integrand, /double)   $
								/(4D*!dpi) /h /(wvl[iw]) *1D-8	
		int[iw,it]		= int[iw,it] + abund(iabund[iz]) *ff[iabund[iz],iw,it]
	  endfor
	endif
    endfor
  endfor
JUMPENDFOR:
endfor


;ergs
if NOT(keyword_set(photons)) then begin
  for it=0, NT-1, 1 do begin
    int[*,it] = int[*,it] *(h*c/wvl[*])
    for iz=0, NZ-1, 1 do begin
      ff[iz,*,it] = ff[iz,*,it] *(h*c/wvl[*])
    endfor
  endfor
endif

;convert back to Angstroms
wvl		= wvl*1D+8

jumpEnd:


message,/info,string('took',systime(1)-t0,' seconds',format='(a,f7.1,a)')
end







;---------------------------------------------------------------------------------------------------------------------
PRO read_gffew, gffew, ei, wi
;
;   read gffew.dat file of R. S. Sutherland, 1998, MNRAS, 300, 321.
;       free-free gaunt factors as function of	scaled initial energy ei  = E_i/(Z^Ry)
;						and scaled photon energy wi = h*nu/(Z^Ry)
;
openr, lur, !data_pth +'/continuum/gffew.dat', /get_lun
;
nei	= 18*3 +1			;number of scaled initial energies
nwi	= 17*3 +1			;number of scaled photon energies
;
str	= ''
gffew	= dblarr(nei,nwi)
ei	= 10^(dindgen(nei)/3D -9D)
wi	= 10^(dindgen(nwi)/3D -8D)
f3	= fltarr(3)
;
for q=0,4 do readf,lur,str		;q is the dummy variable
;
for q=0, 2424, 1 do begin
;
	readf, lur, f3
	f = double(f3)
	i = fix((alog10(f[0])+9D)*3D +0.1D)
	j = fix((alog10(f[1])+8D)*3D +0.1D)
	gffew[i,j] = f[2]
endfor
;
free_lun,lur
;
END