;+
; NAME:
;      K_DR_SUPPRESS
;
; PURPOSE:
;      Computes the total dielectronic recombination rate for the
;      specified ion, including the density-dependent suppression
;      factor as specified by Nikolic et al. (2018, ApJS, 237, 41). 
;
; CATEGORY:
;      CHIANTI; atomic data; recombination.
;
; CALLING SEQUENCE:
;	Result = CH_DR_SUPPRESS( ion_name, temperature )
;
; INPUTS:
;       Ion_Name:  The name of the recombining ion in CHIANTI
;                  format. For example, 'o_5' for O V.
;       Temperature: An array of temperatures (units: K).
;
; OPTIONAL INPUTS:
;       Density:  An electron number density (units: cm^-3) for which
;                 the suppression factor is calculated.
;
; KEYWORD PARAMETERS:
;       NO_LOWT:  If set, then the low-T correction factor is not
;                 applied. See Sect. 2.3 of Nikolic et al.
;       QUIET:    If set, then information is not printed to the IDL
;                 window. 
; 
; OUTPUTS:
;       The dielectronic recombination rate (units: cm^3 s^-1) for the
;       specified recombining ion tabulated for specified temperatures.
;
;
; CALLS:
;       K_DIEL_RECOMB, CONVERTNAME
;
; EXAMPLE:
;       IDL> log_temp=findgen(101)/20.+4.0
;       IDL> rate=ch_dr_suppress('si_4',10.^log_temp)
;
; MODIFICATION HISTORY:
;      Ver.1, 25-Apr-2017, Peter Young
;      Ver.2, 27-Jul-2017, Peter Young
;        If neither pressure or density are provided, then just return
;        the zero-density rates.
;      Ver.3, 7-Nov-2017, Peter Young
;        Updated call to ch_dr_a_n, and added call to ch_dr_exp_eps
;        (for low temperature modification to DR rates).
;      Ver.4, 12-Jun-2020, Peter Young
;        Updated header, changing references to Nikolic et al. (2018)
;        instead of Nikolic et al. (2013); code unchanged. 
;-     Ver for KAPPA 17-Nov-2023  E. Dz.



FUNCTION k_dr_suppress, kappa, gname, temperature, density=density, $
                         quiet=quiet,filename=filename, $
                         pressure=pressure


IF n_params() LT 2 THEN BEGIN
  print,'Use:  IDL> rate=ch_dr_suppress(kappa, gname, temperature [, density=, pressure=, '
  print,'                               /quiet])'
  return,-1
ENDIF 

;
; Get zero density rates
;
rate0=k_diel_recomb(kappa,gname,temperature,quiet=quiet)
IF rate0[0] EQ -1 THEN return,-1

;
; If neither density or pressure have been specified, then just return
; the zero-density rates from ch_diel_recomb.
;
IF n_elements(density) EQ 0 AND n_elements(pressure) EQ 0 THEN return,rate0


nt=n_elements(temperature)

;
; Note that pressure over-rides density if both are input. 
;
IF n_elements(pressure) NE 0 THEN BEGIN
  density=pressure/temperature
  nd=nt
ENDIF ELSE BEGIN
  nd=1
ENDELSE 


;
; Compute activation density
;
convertname,gname,zet,ion
qe=ion-1     ; charge
;N=zet-ion+1  ; iso-electronic sequence number (hydrogen=1)
;theta=temperature/q^2



restore,!data_pth+'/dr_suppression/koef_h+s.sav'  ;,/verbose    tab,kapa,numb,s_kappa
; 1st index  14 - kappa, 2nd index: 0 - a_j (a_j*T) 1 - c_j, 3rd index maxwellians

sf=1.

s_k=string(kappa,format='(f5.1)')
ind=where(s_k eq s_kappa,count)

if count eq 1 then begin
    sf=0.d+0
    pocet=reform(numb(ind)-1)
    a=reform(tab(ind,0,*))
    c=reform(tab(ind,1,*))
      for i=0,pocet(0) do begin
        temp=temperature*a(i)
        get_suppres_fac,zet,qe,temp,density,fac
        sf=sf+c(i)*fac
      endfor
      goto, lab
      
 endif else begin
      
    if (kappa gt 1.7) and (kappa le 3.7) then begin
       sf=0.d+0
       get_coef_hs,kappa,a,c
       num=n_elements(a)
       for i=0,num-1 do begin
          temp=temperature*a(i)       
          get_suppres_fac,zet,qe,temp,density,fac
          sf=sf+c(i)*fac   
       endfor
       goto, lab
       
    endif else begin
       
       print,'Wrong value of kappa!',s_k
       print,'Kappa must be',s_kappa
       sf=1.d0
       
    endelse
    
      
endelse
   
lab:
;
; Return the suppressed rate.
;
return,rate0*sf

END
