
FUNCTION k_rad_recomb,kappa ,gname,temperature, quiet=quiet
;
; CALLING SEQUENCE:
;	Result = K_RAD_RECOMB( Kappa, Akappa, Ion_Name, Temperature )
;
; INPUTS:
;       Kappa: value of kappa 
;       Gname:  The name of the ion in CHIANTI format, e.g., 'fe_13'
;               for Fe XIII.
;       Temperature:  The temperature(s) in K at which the rates are
;                     required. 
;
; KEYWORD PARAMETERS:
;       QUIET:   If set, then no information will be printed to the
;                screen. 
;
; OUTPUTS:
;       The radiative recombination rate in cm^3 s^-1 at the specified
;       temperatures. If a problem occurs, then -1 is returned.
;
; CALLS: ch_rad_recomb 
;
; EXAMPLE:
;       IDL> logt=findgen(41)/10.+4.0
;       IDL> t=10.^t
;       IDL> kappa=2
;       IDL> akappa=gamma(kappa+1.)/(gamma(kappa-0.5)*(kappa-1.5)^1.5)
;       IDL> rate=k_rad_recomb(kappa,akapa,'o_6',t)
;
; MODIFICATION HISTORY:
;       Ver.1, 25-Oct-2016, Peter Young
;          Code extracted from recomb_rate.pro.
;      Ver. 2 for kappa E. Dz. Jan. 2021 
;-



IF n_params() LT 3 THEN BEGIN
  print,'Use:  IDL> rate = k_rad_recomb( Kappa, Ion_Name, Temperature )'
  return,-1.
ENDIF 

; now get radiative recombination rate
;  
nk=n_elements(temperature)
;
te=temperature
  if nk eq 1 then te=[te]
;
rad_rec=fltarr(nk)  
;  
  for i=0,nk-1 do begin 
    tt=te(i)*[1.25893,1.,0.794328]*((kappa-1.5)/(kappa+0.5))  
    rad_r=ch_rad_recomb(gname,tt,/quiet) ;/quiet
    aa=poly_fit(alog(tt),alog(rad_r),1)
    eta=abs(aa(1))
    radr=ch_rad_recomb(gname,te(i),/quiet)
    rad_rec(i)=radr*gamma(kappa+eta-0.5)/(kappa-1.5)^eta/gamma(kappa-0.5)
  endfor
;

return,rad_rec

END
